import textwrap
import unittest
from typing import Tuple, Optional, MutableMapping, List

from icontract import ensure

import tests.common
from aas_core_codegen import intermediate, infer_for_schema
from aas_core_codegen.common import Identifier, Error


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_constraints_by_properties_of_class_something(
    source: str,
) -> Tuple[
    Optional[MutableMapping[intermediate.Property, infer_for_schema.LenConstraint]],
    Optional[List[Error]],
]:
    """Translate the ``source`` into inferred constraints of the class ``Something``."""
    symbol_table, error = tests.common.translate_source_to_intermediate(source=source)
    assert error is None, tests.common.most_underlying_messages(error)
    assert symbol_table is not None
    symbol = symbol_table.must_find(Identifier("Something"))
    assert isinstance(symbol, intermediate.Class)

    result = infer_for_schema.infer_len_constraints_by_class_properties(cls=symbol)

    return result


class Test_expected(unittest.TestCase):
    def test_no_constraints(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual("{}", text)

    def test_min_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 < len(self.some_property)) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=11,
                max_value=None)
            }"""
            ),
            text,
        )

    def test_min_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) > 10) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=11,
                max_value=None)
            }"""
            ),
            text,
        )

    def test_max_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) < 10) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=None,
                max_value=9)
            }"""
            ),
            text,
        )

    def test_max_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 > len(self.some_property)) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=None,
                max_value=9)
            }"""
            ),
            text,
        )

    def test_exact_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 == len(self.some_property)) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=10,
                max_value=10)
            }"""
            ),
            text,
        )

    def test_exact_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) == 10) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=10,
                max_value=10)
            }"""
            ),
            text,
        )

    def test_conditioned_on_property(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self: 
                not (self.some_property is not None)
                or len(self.some_property) == 10
            ) 
            class Something:
                some_property: Optional[str]

                def __init__(self, some_property: Optional[str] = None) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
            {
              'some_property':
              LenConstraint(
                min_value=10,
                max_value=10)
            }"""
            ),
            text,
        )

    def test_inheritance(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) > 3)
            class Parent:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property

                
            @invariant(lambda self: len(self.some_property) > 5) 
            class Something(Parent):
                def __init__(self, some_property: str) -> None:
                    Parent.__init__(
                        self,
                        some_property=some_property
                    ) 


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props, errors = infer_constraints_by_properties_of_class_something(
            source=source
        )

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert by_props is not None

        # NOTE (mristin, 2022-01-02):
        # We infer only the constraints as specified in the class itself, and
        # ignore the constraints of the ancestors in *this particular kind of
        # inference*.

        text = infer_for_schema.dump_len_constraints_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                  'some_property':
                  LenConstraint(
                    min_value=6,
                    max_value=None)
                }"""
            ),
            text,
        )


class Test_unexpected(unittest.TestCase):
    def test_conflicting_min_and_max(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) > 10)
            @invariant(lambda self: len(self.some_property) < 3) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_by_properties_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "The property some_property has conflicting invariants on the length: "
            "the minimum length, 11, contradicts the maximum length 2.",
            tests.common.most_underlying_messages(errors),
        )

    def test_conflicting_min_and_exact(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) > 10)
            @invariant(lambda self: len(self.some_property) == 3) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_by_properties_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "The property some_property has conflicting invariants on the length: "
            "the minimum length, 11, contradicts the exactly expected length 3.",
            tests.common.most_underlying_messages(errors),
        )

    def test_conflicting_max_and_exact(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self.some_property) < 10)
            @invariant(lambda self: len(self.some_property) == 30) 
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_by_properties_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "The property some_property has conflicting invariants on the length: "
            "the maximum length, 9, contradicts the exactly expected length 30.",
            tests.common.most_underlying_messages(errors),
        )


if __name__ == "__main__":
    unittest.main()



"""Represent inferred constraints as strings."""
import io
import textwrap
from typing import Union, Optional, Mapping, Sequence

from aas_core_codegen import stringify, intermediate
from aas_core_codegen.infer_for_schema._len import LenConstraint
from aas_core_codegen.infer_for_schema._pattern import PatternConstraint


def _stringify_len_constraint(
    that: LenConstraint,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("min_value", that.min_value),
            stringify.Property("max_value", that.max_value),
        ],
    )

    return result


def _stringify_pattern_constraint(
    that: PatternConstraint,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("pattern", that.pattern),
        ],
    )

    return result


Dumpable = Union[LenConstraint, PatternConstraint]

_DISPATCH = {
    LenConstraint: _stringify_len_constraint,
    PatternConstraint: _stringify_pattern_constraint,
}

stringify.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def _stringify(that: Optional[Dumpable]) -> Optional[stringify.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify.Entity)
    stringify.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = _stringify(that)
    return stringify.dump(stringified)


def dump_len_constraints_by_properties(
    that: Optional[Mapping[intermediate.Property, LenConstraint]]
) -> str:
    """Represent the map of constraints by properties as strings."""
    if that is None:
        return repr(None)

    if len(that) == 0:
        return "{}"

    writer = io.StringIO()
    writer.write("{\n")
    for i, (prop, dumpable) in enumerate(that.items()):
        writer.write(f"  {prop.name!r}:\n")
        writer.write(textwrap.indent(dump(dumpable), "  "))

        if i < len(that) - 1:
            writer.write(",")

        writer.write("\n")

    writer.write("}")
    return writer.getvalue()


def dump_patterns(that: Optional[Sequence[PatternConstraint]]) -> str:
    """Represent the sequence of patterns as a string."""
    if that is None:
        return repr(None)

    writer = io.StringIO()

    if len(that) == 0:
        writer.write("[]")
    else:
        writer.write("[\n")
        for pattern_i, pattern in enumerate(that):
            writer.write(textwrap.indent(dump(pattern), "  "))

            if pattern_i < len(that) - 1:
                writer.write(",")

            writer.write("\n")

        writer.write("]")

    return writer.getvalue()


def dump_patterns_by_properties(
    that: Optional[Mapping[intermediate.Property, Sequence[PatternConstraint]]]
) -> str:
    """Represent the map of patterns by properties as strings."""
    if that is None:
        return repr(None)

    if len(that) == 0:
        return "{}"

    writer = io.StringIO()
    writer.write("{\n")
    for i, (prop, patterns) in enumerate(that.items()):
        writer.write(f"  {prop.name!r}:\n")
        writer.write(textwrap.indent(dump_patterns(patterns), "  "))

        if i < len(that) - 1:
            writer.write(",")

        writer.write("\n")

    writer.write("}")
    return writer.getvalue()



"""Generate C# code to handle asset administration shells based on the meta-model."""
from typing import TextIO

from aas_core_codegen import specific_implementations, run, intermediate
from aas_core_codegen.csharp import (
    common as csharp_common,
    structure as csharp_structure,
    visitation as csharp_visitation,
    verification as csharp_verification,
    stringification as csharp_stringification,
    jsonization as csharp_jsonization,
)


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    verified_ir_table, errors = csharp_structure.verify(
        symbol_table=context.symbol_table
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of C# code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    namespace_key = specific_implementations.ImplementationKey("namespace.txt")
    namespace_text = context.spec_impls.get(namespace_key, None)
    if namespace_text is None:
        stderr.write(f"The namespace snippet is missing: {namespace_key}\n")
        return 1

    if not csharp_common.NAMESPACE_IDENTIFIER_RE.fullmatch(namespace_text):
        stderr.write(
            f"The text from the snippet {namespace_key} "
            f"is not a valid namespace identifier: {namespace_text!r}\n"
        )
        return 1

    namespace = csharp_common.NamespaceIdentifier(namespace_text)

    # region Structure

    code, errors = csharp_structure.generate(
        symbol_table=verified_ir_table,
        namespace=namespace,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the structures in the C# code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "types.cs"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the C# structures to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Visitation

    code, errors = csharp_visitation.generate(
        symbol_table=context.symbol_table, namespace=namespace
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the C# code for visitation "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "visitation.cs"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the visitation C# code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Verification

    verify_errors = csharp_verification.verify(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if errors is not None:
        run.write_error_report(
            message="Failed to verify the C#-specific structures",
            errors=verify_errors,
            stderr=stderr,
        )
        return 1

    code, errors = csharp_verification.generate(
        symbol_table=verified_ir_table,
        namespace=namespace,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the verification C# code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "verification.cs"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the verification C# code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Stringification

    code, errors = csharp_stringification.generate(
        symbol_table=context.symbol_table, namespace=namespace
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the stringification C# code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "stringification.cs"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the stringification C# code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Jsonization

    code, errors = csharp_jsonization.generate(
        symbol_table=context.symbol_table,
        namespace=namespace,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the jsonization C# code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "jsonization.cs"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the jsonization C# code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



import textwrap
import unittest

import aas_core_codegen.intermediate._hierarchy as _hierarchy
import tests.common
from aas_core_codegen.common import Identifier


class Test_ontology_ok(unittest.TestCase):
    def test_no_ancestors(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                class Something:
                    pass
                    
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )

        assert error is None, f"{error=}"
        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is None, f"{errors=}"
        assert ontology is not None

        ancestors = ontology.list_ancestors(
            cls=symbol_table.must_find_class(Identifier("Something"))
        )

        ancestor_names = [cls.name for cls in ancestors]
        self.assertListEqual([], ancestor_names)

    def test_complex_graph(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                @abstract
                class Another_grand_parent:
                    pass

                @abstract
                class Grand_parent:
                    pass

                @abstract
                class Parent(Grand_parent, Another_grand_parent):
                    pass

                @abstract
                class Another_parent(Grand_parent, Another_grand_parent):
                    pass

                class Something(Parent, Another_parent):
                    pass
                    
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )

        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is None, f"{errors=}"
        assert ontology is not None

        ancestors = ontology.list_ancestors(
            cls=symbol_table.must_find_class(Identifier("Something"))
        )

        ancestor_names = [cls.name for cls in ancestors]

        self.assertListEqual(
            [
                "Another_grand_parent",
                "Grand_parent",
                "Another_parent",
                "Another_grand_parent",
                "Grand_parent",
                "Parent",
            ],
            ancestor_names,
        )


class Test_ontology_fail(unittest.TestCase):
    def test_duplicate_properties_in_ancestors(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                @abstract
                class Abstract:
                    x: int

                @abstract
                class Something(Abstract):
                    x: int
                    
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )
        assert error is None, f"{error=}"
        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is not None and len(errors) == 1

        self.assertEqual(
            "The property has already been defined "
            "in the ancestor class Abstract: x",
            tests.common.most_underlying_messages(errors[0]),
        )

    def test_duplicate_methods_in_ancestors(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                @abstract
                class Abstract:
                    def do_something(self) -> None:
                        pass

                @abstract
                class Something(Abstract):
                    def do_something(self) -> None:
                        pass
                        
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )
        assert error is None, f"{error=}"
        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is not None and len(errors) == 1

        self.assertEqual(
            "The method has already been defined "
            "in the ancestor class Abstract: do_something",
            tests.common.most_underlying_messages(errors[0]),
        )

    def test_missing_constructor_when_the_parent_has_one(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                @abstract
                class Abstract:
                    def __init__(self, x: int) -> None:
                        pass

                class Something(Abstract):
                    pass
                    
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )
        assert error is None, f"{error=}"
        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is not None and len(errors) == 1

        self.assertEqual(
            "The class Something does not specify a constructor, "
            "but the ancestor class Abstract specifies a constructor "
            "with arguments: self, x",
            tests.common.most_underlying_messages(errors[0]),
        )

    def test_cycle_inheritance(self) -> None:
        symbol_table, error = tests.common.parse_source(
            textwrap.dedent(
                """\
                @abstract
                class Cycle(Something):
                    pass

                @abstract
                class Something(Cycle):
                    pass
                    
                class Reference:
                    pass
                
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                """
            )
        )
        assert error is None, f"{error=}"
        assert symbol_table is not None

        ontology, errors = _hierarchy.map_symbol_table_to_ontology(symbol_table)
        assert errors is not None and len(errors) == 1

        self.assertEqual(
            "Expected no cycles in the inheritance, "
            "but the class Cycle has been observed in a cycle",
            tests.common.most_underlying_messages(errors[0]),
        )


if __name__ == "__main__":
    unittest.main()



"""Generate the invariant verifiers from the intermediate representation."""
import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Union,
    Final,
    Set,
    MutableMapping,
    Mapping,
)

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import Error, Stripped, assert_never, Identifier
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
    unrolling as csharp_unrolling,
    description as csharp_description,
)
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II, INDENT3 as III
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree


# region Verify


def verify(
    spec_impls: specific_implementations.SpecificImplementations,
    verification_functions: Sequence[intermediate.Verification],
) -> Optional[List[str]]:
    """Verify all the implementation snippets related to verification."""
    errors = []  # type: List[str]

    expected_keys = [
        specific_implementations.ImplementationKey("Verification/Error.cs"),
        specific_implementations.ImplementationKey("Verification/Errors.cs"),
    ]

    for func in verification_functions:
        if isinstance(func, intermediate.ImplementationSpecificVerification):
            expected_keys.append(
                specific_implementations.ImplementationKey(
                    f"Verification/{func.name}.cs"
                ),
            )

    for key in expected_keys:
        if key not in spec_impls:
            errors.append(f"The implementation snippet is missing for: {key}")

    if len(errors) == 0:
        return None

    return errors


# endregion

# region Generate


class _PatternVerificationTranspiler(parse_tree.RestrictedTransformer[Stripped]):
    """Transpile a statement of a pattern verification into C#."""

    def __init__(self, defined_variables: Set[Identifier]) -> None:
        """
        Initialize with the given values.

        The ``initialized_variables`` are shared between different statement
        transpilations. It is also mutated when assignments are transpiled. We need to
        keep track of variables so that we know when we have to define them, and when
        we can simply assign them a value, if they have been already defined.
        """
        self.defined_variables = defined_variables

    def transform_constant(self, node: parse_tree.Constant) -> Stripped:
        if isinstance(node.value, str):
            return Stripped(csharp_common.string_literal(node.value))
        else:
            raise AssertionError(f"Unexpected {node=}")

    def transform_name(self, node: parse_tree.Name) -> Stripped:
        return Stripped(csharp_naming.variable_name(node.identifier))

    def transform_joined_str(self, node: parse_tree.JoinedStr) -> Stripped:
        parts = []  # type: List[str]
        for value in node.values:
            if isinstance(value, str):
                string_literal = csharp_common.string_literal(
                    value.replace("{", "{{").replace("}", "}}")
                )

                # We need to remove double-quotes since we are joining everything
                # ourselves later.

                assert string_literal.startswith('"') and string_literal.endswith('"')

                string_literal_wo_quotes = string_literal[1:-1]
                parts.append(string_literal_wo_quotes)

            elif isinstance(value, parse_tree.FormattedValue):
                code = self.transform(value.value)
                assert (
                    "\n" not in code
                ), f"New-lines are not expected in formatted values, but got: {code}"

                parts.append(f"{{{code}}}")
            else:
                assert_never(value)

        writer = io.StringIO()
        writer.write('$"')
        for part in parts:
            writer.write(part)

        writer.write('"')

        return Stripped(writer.getvalue())

    def transform_assignment(self, node: parse_tree.Assignment) -> Stripped:
        assert isinstance(node.target, parse_tree.Name)
        variable = csharp_naming.variable_name(node.target.identifier)
        code = self.transform(node.value)

        if node.target.identifier in self.defined_variables:
            return Stripped(f"{variable} = {code};")

        else:
            self.defined_variables.add(node.target.identifier)
            return Stripped(f"var {variable} = {code};")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_pattern_verification(
    verification: intermediate.PatternVerification,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verification function that checks the regular expressions."""
    # NOTE (mristin, 2021-12-19):
    # We assume that we performed all the checks at the intermediate stage.

    construct_name = csharp_naming.private_method_name(
        Identifier(f"construct_{verification.name}")
    )

    blocks = []  # type: List[Stripped]

    # region Construct block

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            f"""\
        private static Regex {construct_name}()
        {{
        """
        )
    )

    defined_variables = set()  # type: Set[Identifier]
    transpiler = _PatternVerificationTranspiler(defined_variables=defined_variables)

    for i, stmt in enumerate(verification.parsed.body):
        if i == len(verification.parsed.body) - 1:
            break

        code = transpiler.transform(stmt)
        writer.write(textwrap.indent(code, I))
        writer.write("\n")

    if len(verification.parsed.body) >= 2:
        writer.write("\n")

    assert len(verification.parsed.body) >= 1

    assert isinstance(verification.parsed.body[-1], parse_tree.Return)
    # noinspection PyUnresolvedReferences
    assert isinstance(verification.parsed.body[-1].value, parse_tree.IsNotNone)
    # noinspection PyUnresolvedReferences
    assert isinstance(verification.parsed.body[-1].value.value, parse_tree.FunctionCall)
    # noinspection PyUnresolvedReferences
    assert verification.parsed.body[-1].value.value.name.identifier == "match"

    # noinspection PyUnresolvedReferences
    match_call = verification.parsed.body[-1].value.value

    assert isinstance(
        match_call, parse_tree.FunctionCall
    ), f"{parse_tree.dump(match_call)}"
    assert match_call.name.identifier == "match"

    assert isinstance(match_call.args[0], parse_tree.Expression)
    pattern_expr = transpiler.transform(match_call.args[0])

    # A pragmatic heuristics for breaking lines
    if len(pattern_expr) < 50:
        writer.write(textwrap.indent(f"return new Regex({pattern_expr});\n", I))
    else:
        writer.write(textwrap.indent(f"return new Regex(\n{I}{pattern_expr});\n", I))

    writer.write("}")

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region Initialize the regex

    regex_name = csharp_naming.private_property_name(
        Identifier(f"regex_{verification.name}")
    )

    blocks.append(
        Stripped(f"private static readonly Regex {regex_name} = {construct_name}();")
    )

    assert len(verification.arguments) == 1
    assert isinstance(
        verification.arguments[0].type_annotation, intermediate.PrimitiveTypeAnnotation
    )
    # noinspection PyUnresolvedReferences
    assert (
        verification.arguments[0].type_annotation.a_type
        == intermediate.PrimitiveType.STR
    )

    arg_name = csharp_naming.argument_name(verification.arguments[0].name)

    writer = io.StringIO()
    if verification.description is not None:
        comment, error = csharp_description.generate_comment(verification.description)
        if error is not None:
            return None, error

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    writer.write(
        textwrap.dedent(
            f"""\
            public static bool IsMimeType(string {arg_name})
            {{
            {I}return {regex_name}.IsMatch({arg_name});
            }}"""
        )
    )

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    return Stripped(writer.getvalue()), None


def _generate_enum_value_sets(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate a class that pre-computes the sets of allowed enumeration literals."""
    blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if not isinstance(symbol, intermediate.Enumeration):
            continue

        enum_name = csharp_naming.enum_name(symbol.name)

        if len(symbol.literals) == 0:
            blocks.append(
                Stripped(
                    f"public static HashSet<int> For{enum_name} = new HashSet<int>();"
                )
            )
        else:
            hash_set_writer = io.StringIO()
            hash_set_writer.write(
                f"public static HashSet<int> For{enum_name} = new HashSet<int>\n{{\n"
            )

            for i, literal in enumerate(symbol.literals):
                literal_name = csharp_naming.enum_literal_name(literal.name)
                hash_set_writer.write(f"{I}(int)Aas.{enum_name}.{literal_name}")
                if i < len(symbol.literals) - 1:
                    hash_set_writer.write(",\n")
                else:
                    hash_set_writer.write("\n")

            hash_set_writer.write("};")

            blocks.append(Stripped(hash_set_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            """\
        /// <summary>
        /// Hash allowed enum values for efficient validation of enums.
        /// </summary>
        private static class EnumValueSet
        {
        """
        )
    )
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // private static class EnumValueSet")

    return Stripped(writer.getvalue())


class _EnumerationCheckUnroller(csharp_unrolling.Unroller):
    #: Symbol to be used to represent references within an AAS
    _ref_association: Final[intermediate.ClassUnion]

    def __init__(self, ref_association: intermediate.ClassUnion) -> None:
        """Initialize with the given values."""
        self._ref_association = ref_association

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        # Primitives are not enumerations, so nothing to check here.
        return []

    # noinspection PyUnusedLocal
    def _unroll_our_type_or_ref_annotation(
        self,
        unrollee_expr: str,
        type_annotation: Union[
            intermediate.OurTypeAnnotation, intermediate.RefTypeAnnotation
        ],
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """
        Generate the code for both our atomic type annotations and references.

        We merged :py:method:`._unroll_our_type_annotation` and
        :py:method:`._unroll_ref_type_annotation` together since they differ in only
        which symbol is unrolled over.
        """
        symbol = None  # type: Optional[intermediate.Symbol]
        if isinstance(type_annotation, intermediate.OurTypeAnnotation):
            symbol = type_annotation.symbol
        elif isinstance(type_annotation, intermediate.RefTypeAnnotation):
            symbol = self._ref_association
        else:
            assert_never(type_annotation)

        assert symbol is not None

        if not isinstance(symbol, intermediate.Enumeration):
            # NOTE (mristin, 2021-12-25):
            # We do not descend into other types as this is generating the code
            # only for the non-descend case.
            return []

        enum_name = csharp_naming.enum_name(symbol.name)
        joined_pth = "/".join(path)

        return [
            csharp_unrolling.Node(
                text=textwrap.dedent(
                    f"""\
                if (!Verification.Implementation.EnumValueSet.For{enum_name}.Contains(
                {II}(int){unrollee_expr}))
                {{
                {I}errors.Add(
                {II}new Verification.Error(
                {III}$"{{path}}/{joined_pth}",
                {III}$"Invalid {{nameof(Aas.{enum_name})}}: {{{unrollee_expr}}}"));
                }}"""
                ),
                children=[],
            )
        ]

    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        return self._unroll_our_type_or_ref_annotation(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        # Descend into the list items
        item_var = csharp_unrolling.Unroller._loop_var_name(
            level=item_level, suffix="Item"
        )

        children = self.unroll(
            unrollee_expr=item_var,
            type_annotation=type_annotation.items,
            path=path + [f"{{{item_var}}}"],
            item_level=item_level + 1,
            key_value_level=key_value_level,
        )

        if len(children) == 0:
            return []

        node = csharp_unrolling.Node(
            text=f"foreach (var {item_var} in {unrollee_expr})", children=children
        )

        return [node]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )

        if len(children) > 0:
            return [
                csharp_unrolling.Node(
                    text=f"if ({unrollee_expr} != null)", children=children
                )
            ]
        else:
            return []

    def _unroll_ref_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.RefTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        return self._unroll_our_type_or_ref_annotation(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )


def _unroll_enumeration_check(
    prop: intermediate.Property, ref_association: intermediate.ClassUnion
) -> Stripped:
    """
    Generate the code for unrolling the enumeration checks for the given property.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    prop_name = csharp_naming.property_name(prop.name)

    unroller = _EnumerationCheckUnroller(ref_association=ref_association)

    roots = unroller.unroll(
        unrollee_expr=f"that.{prop_name}",
        type_annotation=prop.type_annotation,
        path=[prop_name],
        item_level=0,
        key_value_level=0,
    )

    if len(roots) == 0:
        return Stripped("")

    blocks = [csharp_unrolling.render(root) for root in roots]
    return Stripped("\n\n".join(blocks))


class _ConstrainedPrimitiveCheckUnroller(csharp_unrolling.Unroller):
    #: Symbol to be used to represent references within an AAS
    _ref_association: Final[intermediate.Class]

    def __init__(self, ref_association: intermediate.ClassUnion) -> None:
        """Initialize with the given values."""
        self._ref_association = ref_association

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        # Nothing to unroll for primitives.
        return []

    # noinspection PyUnusedLocal
    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """
        Generate the code for both our atomic type annotations and references.

        We merged :py:method:`._unroll_our_type_annotation` and
        :py:method:`._unroll_ref_type_annotation` together since they differ in only
        which symbol is unrolled over.
        """
        if not isinstance(type_annotation.symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2021-12-25):
            # We do not descend into other types as this is generating the code
            # only for the non-descend case. The recursive verifier will descend into
            # properties of other classes.
            return []

        cls_name = csharp_naming.class_name(type_annotation.symbol.name)

        joined_pth = "/".join(path)

        return [
            csharp_unrolling.Node(
                text=textwrap.dedent(
                    f"""\
                    Verification.Implementation.Verify{cls_name}(
                    {I}{unrollee_expr},
                    {I}$"{{path}}/{joined_pth}",
                    {I}errors);"""
                ),
                children=[],
            )
        ]

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        # Descend into the list items
        item_var = csharp_unrolling.Unroller._loop_var_name(
            level=item_level, suffix="Item"
        )

        children = self.unroll(
            unrollee_expr=item_var,
            type_annotation=type_annotation.items,
            path=path + [f"{{{item_var}}}"],
            item_level=item_level + 1,
            key_value_level=key_value_level,
        )

        if len(children) == 0:
            return []

        node = csharp_unrolling.Node(
            text=f"foreach (var {item_var} in {unrollee_expr})", children=children
        )

        return [node]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )

        if len(children) > 0:
            return [
                csharp_unrolling.Node(
                    text=f"if ({unrollee_expr} != null)", children=children
                )
            ]
        else:
            return []

    def _unroll_ref_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.RefTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        assert isinstance(
            self._ref_association, intermediate.Class
        ), "Assume that the ``ref_association`` is a class so that we don't unroll"

        # We do not descend into classes, this is done in the recursive verifier.
        return []


def _unroll_constrained_primitive_check(
    prop: intermediate.Property, ref_association: intermediate.ClassUnion
) -> Stripped:
    """
    Generate the code for unrolling checking primitive constraints on the property.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    prop_name = csharp_naming.property_name(prop.name)

    unroller = _ConstrainedPrimitiveCheckUnroller(ref_association=ref_association)

    roots = unroller.unroll(
        unrollee_expr=f"that.{prop_name}",
        type_annotation=prop.type_annotation,
        path=[prop_name],
        item_level=0,
        key_value_level=0,
    )

    if len(roots) == 0:
        return Stripped("")

    blocks = [csharp_unrolling.render(root) for root in roots]
    return Stripped("\n\n".join(blocks))


class _InvariantTranspiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile an invariant expression into a code, or an error."""

    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: Mapping[
            Identifier, intermediate_type_inference.TypeAnnotationUnion
        ],
    ) -> None:
        """Initialize with the given values."""
        self.type_map = type_map
        self.environment = environment

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_member(
        self, node: parse_tree.Member
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        instance, error = self.transform(node.instance)
        if error is not None:
            return None, error

        instance_type = self.type_map[node.instance]
        # Ignore optionals as they need to be checked before in the code
        while isinstance(
            instance_type, intermediate_type_inference.OptionalTypeAnnotation
        ):
            instance_type = instance_type.value

        member_type = self.type_map[node]
        while isinstance(
            member_type, intermediate_type_inference.OptionalTypeAnnotation
        ):
            member_type = member_type.value

        member_name = None  # type: Optional[str]

        if isinstance(
            instance_type, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type.symbol, intermediate.Enumeration):
            # The member denotes a literal of an enumeration.
            member_name = csharp_naming.enum_literal_name(node.name)

        elif isinstance(member_type, intermediate_type_inference.MethodTypeAnnotation):
            member_name = csharp_naming.method_name(node.name)

        elif isinstance(
            instance_type, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type.symbol, intermediate.Class):
            if node.name in instance_type.symbol.properties_by_name:
                member_name = csharp_naming.property_name(node.name)
            else:
                return None, Error(
                    node.original_node,
                    f"The property {node.name!r} has not been defined "
                    f"in the class {instance_type.symbol.name!r}",
                )

        else:
            return None, Error(
                node.original_node,
                f"We do not know how to generate the member access. The inferred type "
                f"of the instance was {instance_type}, while the member type "
                f"was {member_type}. However, we do not know how to resolve "
                f"the member {node.name!r} in {instance_type}.",
            )

        assert member_name is not None

        return Stripped(f"{instance}.{member_name}"), None

    _CSHARP_COMPARISON_MAP = {
        parse_tree.Comparator.LT: "<",
        parse_tree.Comparator.LE: "<=",
        parse_tree.Comparator.GT: ">",
        parse_tree.Comparator.GE: ">=",
        parse_tree.Comparator.EQ: "==",
        parse_tree.Comparator.NE: "!=",
    }

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_comparison(
        self, node: parse_tree.Comparison
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        comparator = _InvariantTranspiler._CSHARP_COMPARISON_MAP[node.op]

        errors = []

        left, error = self.transform(node.left)
        if error is not None:
            errors.append(error)

        right, error = self.transform(node.right)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the comparison", errors
            )

        no_parentheses_types = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.Constant,
        )

        if isinstance(node.left, no_parentheses_types) and isinstance(
            node.right, no_parentheses_types
        ):
            return Stripped(f"{left} {comparator} {right}"), None

        return Stripped(f"({left}) {comparator} ({right})"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_implication(
        self, node: parse_tree.Implication
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []

        antecedent, error = self.transform(node.antecedent)
        if error is not None:
            errors.append(error)

        consequent, error = self.transform(node.consequent)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the implication", errors
            )

        assert antecedent is not None
        assert consequent is not None

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
        )

        if isinstance(node.antecedent, no_parentheses_types_in_this_context):
            not_antecedent = f"!{antecedent}"
        else:
            not_antecedent = f"!({antecedent})"

        if not isinstance(node.consequent, no_parentheses_types_in_this_context):
            consequent = Stripped(f"({consequent})")

        return Stripped(f"{not_antecedent}\n" f"|| {consequent}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_method_call(
        self, node: parse_tree.MethodCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        instance, error = self.transform(node.member.instance)
        if error is not None:
            errors.append(error)

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg, error = self.transform(arg_node)
            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the method call", errors
            )

        assert instance is not None

        if not isinstance(node.member.instance, (parse_tree.Name, parse_tree.Member)):
            instance = Stripped(f"({instance})")

        method_name = csharp_naming.method_name(node.member.name)

        joined_args = ", ".join(args)

        # Apply heuristic for breaking the lines
        if len(joined_args) > 50:
            writer = io.StringIO()
            writer.write(f"{instance}.{method_name}(\n")

            for i, arg in enumerate(args):
                writer.write(f"{I}{arg}")

                if i == len(args) - 1:
                    writer.write(")")
                else:
                    writer.write(",\n")

            return Stripped(writer.getvalue()), None
        else:
            return Stripped(f"{instance}.{method_name}({joined_args})"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_function_call(
        self, node: parse_tree.FunctionCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg, error = self.transform(arg_node)
            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the function call", errors
            )

        # NOTE (mristin, 2021-12-16):
        # The validity of the arguments is checked in
        # :py:func:`aas_core_codegen.intermediate._translate.translate`, so we do not
        # have to test for argument arity here.

        func_type = self.type_map[node.name]

        if not isinstance(
            func_type, intermediate_type_inference.FunctionTypeAnnotationUnionAsTuple
        ):
            return None, Error(
                node.name.original_node,
                f"Expected the name to refer to a function, "
                f"but its inferred type was {func_type}",
            )

        if isinstance(
            func_type, intermediate_type_inference.VerificationTypeAnnotation
        ):
            method_name = csharp_naming.method_name(func_type.func.name)

            joined_args = ", ".join(args)

            # Apply heuristic for breaking the lines
            if len(joined_args) > 50:
                writer = io.StringIO()
                writer.write(f"Verification.{method_name}(\n")

                for i, arg in enumerate(args):
                    writer.write(f"{I}{arg}")

                    if i == len(args) - 1:
                        writer.write(")")
                    else:
                        writer.write(",\n")

                return Stripped(writer.getvalue()), None
            else:
                return Stripped(f"Verification.{method_name}({joined_args})"), None

        elif isinstance(
            func_type, intermediate_type_inference.BuiltinFunctionTypeAnnotation
        ):
            if func_type.func.name == "len":
                assert len(args) == 1, (
                    f"Expected exactly one argument, but got: {args}; "
                    f"this should have been caught before."
                )

                collection_node = node.args[0]
                if not isinstance(
                    collection_node,
                    (parse_tree.Name, parse_tree.Member, parse_tree.MethodCall),
                ):
                    collection = f"({args[0]})"
                else:
                    collection = args[0]

                arg_type = self.type_map[node.args[0]]
                while isinstance(
                    arg_type, intermediate_type_inference.OptionalTypeAnnotation
                ):
                    arg_type = arg_type.value

                if (
                    isinstance(
                        arg_type, intermediate_type_inference.PrimitiveTypeAnnotation
                    )
                    and arg_type.a_type == intermediate_type_inference.PrimitiveType.STR
                ):
                    return Stripped(f"{collection}.Length"), None

                elif (
                    isinstance(arg_type, intermediate_type_inference.OurTypeAnnotation)
                    and isinstance(arg_type.symbol, intermediate.ConstrainedPrimitive)
                    and arg_type.symbol.constrainee == intermediate.PrimitiveType.STR
                ):
                    return Stripped(f"{collection}.Length"), None

                elif isinstance(
                    arg_type, intermediate_type_inference.ListTypeAnnotation
                ):
                    return Stripped(f"{collection}.Count"), None

                else:
                    return None, Error(
                        node.original_node,
                        f"We do not know how to compute the length on type {arg_type}",
                        errors,
                    )
            else:
                return None, Error(
                    node.original_node,
                    f"The handling of the built-in function {node.name!r} has not "
                    f"been implemented",
                )
        else:
            assert_never(func_type)

        raise AssertionError("Should not have gotten here")

    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, bool):
            return Stripped("true" if node.value else "false"), None
        elif isinstance(node.value, (int, float)):
            return Stripped(str(node.value)), None
        elif isinstance(node.value, str):
            return Stripped(repr(node.value)), None
        else:
            assert_never(node.value)

        raise AssertionError("Should not have gotten here")

    def transform_is_none(
        self, node: parse_tree.IsNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )
        if isinstance(node.value, no_parentheses_types):
            return Stripped(f"{value} == null"), None
        else:
            return Stripped(f"({value}) == null"), None

    def transform_is_not_none(
        self, node: parse_tree.IsNotNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types_in_this_context = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )
        if isinstance(node.value, no_parentheses_types_in_this_context):
            return Stripped(f"{value} != null"), None
        else:
            return Stripped(f"({value}) != null"), None

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier == "self":
            # The ``that`` refers to the argument of the verification function.
            return Stripped("that"), None

        return Stripped(csharp_naming.variable_name(node.identifier)), None

    def transform_and(
        self, node: parse_tree.And
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]
        values = []  # type: List[Stripped]

        for value_node in node.values:
            value, error = self.transform(value_node)
            if error is not None:
                errors.append(error)
                continue

            assert value is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Comparison,
                parse_tree.Name,
            )

            if not isinstance(value_node, no_parentheses_types_in_this_context):
                value = Stripped(f"({value})")

            values.append(value)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the conjunction", errors
            )

        # BEFORE-RELEASE (mristin, 2021-12-13):
        #  add heuristic for breaking the lines
        return Stripped(" && ".join(values)), None

    def transform_or(
        self, node: parse_tree.Or
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]
        values = []  # type: List[Stripped]

        for value_node in node.values:
            value, error = self.transform(value_node)
            if error is not None:
                errors.append(error)
                continue

            assert value is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Comparison,
                parse_tree.Name,
            )

            if not isinstance(value_node, no_parentheses_types_in_this_context):
                value = Stripped(f"({value})")

            values.append(value)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the conjunction", errors
            )

        # BEFORE-RELEASE (mristin, 2021-12-13):
        #  add heuristic for breaking the lines
        return Stripped(" || ".join(values)), None

    def transform_declaration(
        self, node: parse_tree.Declaration
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        # BEFORE-RELEASE (mristin, 2021-12-13):
        #  implement once we got to end-to-end with serialization
        raise NotImplementedError()

    def transform_expression_with_declarations(
        self, node: parse_tree.ExpressionWithDeclarations
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        # BEFORE-RELEASE (mristin, 2021-12-13):
        #  implement once we got to end-to-end with serialization
        raise NotImplementedError()

    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        parts = []  # type: List[str]
        for value in node.values:
            if isinstance(value, str):
                string_literal = csharp_common.string_literal(
                    value.replace("{", "{{").replace("}", "}}")
                )

                # We need to remove double-quotes since we are joining everything
                # ourselves later.

                assert string_literal.startswith('"') and string_literal.endswith('"')

                string_literal_wo_quotes = string_literal[1:-1]
                parts.append(string_literal_wo_quotes)

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self.transform(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                assert (
                    "\n" not in code
                ), f"New-lines are not expected in formatted values, but got: {code}"

                parts.append(f"{{{code}}}")
            else:
                assert_never(value)

        writer = io.StringIO()
        writer.write('$"')
        for part in parts:
            writer.write(part)

        writer.write('"')

        return Stripped(writer.getvalue()), None


# noinspection PyProtectedMember,PyProtectedMember
assert all(
    op in _InvariantTranspiler._CSHARP_COMPARISON_MAP for op in parse_tree.Comparator
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_invariant(
    invariant: intermediate.Invariant,
    symbol_table: intermediate.SymbolTable,
    environment: Mapping[Identifier, intermediate_type_inference.TypeAnnotationUnion],
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Translate the invariant from the meta-model into C# code."""
    # NOTE (mristin, 2021-10-24):
    # We manually transpile the invariant from our custom syntax without additional
    # semantic analysis in the :py:mod:`aas_core_codegen.intermediate` layer.
    #
    # While this might seem repetitive ("unDRY"), we are still not sure about
    # the appropriate abstraction. After we implement the code generation for a couple
    # of languages, we hope to have a much better understanding about the necessary
    # abstractions.

    type_inferrer = intermediate_type_inference.Inferrer(
        symbol_table=symbol_table, environment=environment
    )
    _ = type_inferrer.transform(invariant.body)
    if len(type_inferrer.errors):
        return None, Error(
            invariant.parsed.node,
            "Failed to infer the types in the invariant",
            type_inferrer.errors,
        )

    transformer = _InvariantTranspiler(
        type_map=type_inferrer.type_map, environment=environment
    )
    expr, error = transformer.transform(invariant.parsed.body)
    if error is not None:
        return None, error

    assert expr is not None

    writer = io.StringIO()
    if len(expr) > 50 or "\n" in expr:
        writer.write("if (!(\n")
        writer.write(textwrap.indent(expr, I))
        writer.write("))\n{\n")
    else:
        no_parenthesis_type_in_this_context = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )

        if isinstance(invariant.parsed.body, no_parenthesis_type_in_this_context):
            not_expr = f"!{expr}"
        else:
            not_expr = f"!({expr})"

        writer.write(f"if ({not_expr})\n{{\n")

    writer.write(
        textwrap.indent(
            textwrap.dedent(
                f"""\
        errors.Add(
        {I}new Verification.Error(
        {II}path,
        {II}"Invariant violated:\\n" +
        """
            ),
            I,
        )
    )

    lines = []  # type: List[str]
    if invariant.description is not None:
        lines = invariant.description.splitlines()

    lines = lines + expr.splitlines()

    for i, line in enumerate(lines):
        if i < len(lines) - 1:
            line_literal = csharp_common.string_literal(line + "\n")
            writer.write(f"{III}{line_literal} +\n")
        else:
            writer.write(f"{III}{csharp_common.string_literal(line)}));")

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_implementation_verify(
    something: Union[intermediate.ConcreteClass, intermediate.ConstrainedPrimitive],
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verify function in the ``Implementation`` class."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    # Build up the environment;
    # see https://craftinginterpreters.com/resolving-and-binding.html
    environment: MutableMapping[
        Identifier, intermediate_type_inference.TypeAnnotationUnion
    ] = {
        Identifier("len"): intermediate_type_inference.BuiltinFunctionTypeAnnotation(
            func=intermediate_type_inference.BuiltinFunction(
                name=Identifier("len"),
                returns=intermediate_type_inference.PrimitiveTypeAnnotation(
                    intermediate_type_inference.PrimitiveType.LENGTH
                ),
            )
        )
    }

    for verification in symbol_table.verification_functions:
        assert verification.name not in environment
        environment[
            verification.name
        ] = intermediate_type_inference.VerificationTypeAnnotation(func=verification)

    assert "self" not in environment
    environment[Identifier("self")] = intermediate_type_inference.OurTypeAnnotation(
        symbol=something
    )

    for invariant in something.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(error)
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, Error(
            something.parsed.node,
            f"Failed to parse one or more invariants of the class {something.name!r}",
            errors,
        )

    if isinstance(something, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(something, intermediate.ConcreteClass):
        for prop in something.properties:
            enum_check_block = _unroll_enumeration_check(
                prop=prop, ref_association=symbol_table.ref_association
            )
            if enum_check_block != "":
                blocks.append(Stripped("if (errors.Full()) return;"))
                blocks.append(enum_check_block)

            constrained_primitive_check = _unroll_constrained_primitive_check(
                prop=prop, ref_association=symbol_table.ref_association
            )
            if constrained_primitive_check != "":
                blocks.append(Stripped("if (errors.Full()) return;"))
                blocks.append(constrained_primitive_check)
    else:
        assert_never(something)

    if len(blocks) == 0:
        blocks.append(Stripped("// There is no verification specified."))

    cls_name = csharp_naming.class_name(something.name)

    that_type = None  # type: Optional[str]
    if isinstance(something, intermediate.ConstrainedPrimitive):
        that_type = csharp_common.PRIMITIVE_TYPE_MAP[something.constrainee]
    elif isinstance(something, intermediate.ConcreteClass):
        that_type = f"Aas.{cls_name}"
    else:
        assert_never(something)

    assert that_type is not None

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            f"""\
        /// <summary>
        /// Verify <paramref name="that" /> and append any errors to
        /// <paramref name="Errors" />.
        ///
        /// The <paramref name="path" /> localizes <paramref name="that" />.
        /// </summary>
        public static void Verify{cls_name} (
        {I}{that_type} that,
        {I}string path,
        {I}Verification.Errors errors)
        {{
        """
        )
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    assert len(errors) == 0
    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_implementation_class(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a private static class, ``Implementation``, with verification logic."""
    errors = []  # type: List[Error]
    blocks = [
        _generate_enum_value_sets(symbol_table=symbol_table)
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            implementation_verify, error = _generate_implementation_verify(
                something=symbol, symbol_table=symbol_table
            )
            if error is not None:
                errors.append(error)
                continue

            assert implementation_verify is not None

            if implementation_verify != "":
                blocks.append(implementation_verify)

        elif isinstance(symbol, intermediate.AbstractClass):
            # No verification of interfaces, and all abstract classes are modeled as
            # interfaces in C#.
            continue

        elif isinstance(symbol, intermediate.ConcreteClass):
            if symbol.is_implementation_specific:
                verify_key = specific_implementations.ImplementationKey(
                    f"Verification/Implementation/verify_{symbol.name}.cs"
                )
                if verify_key not in spec_impls:
                    errors.append(
                        Error(
                            symbol.parsed.node,
                            f"The implementation snippet is missing for "
                            f"the ``Verify`` method "
                            f"of the ``Verification.Implementation`` class: {verify_key}",
                        )
                    )
                    continue

                blocks.append(spec_impls[verify_key])
            else:
                implementation_verify, error = _generate_implementation_verify(
                    something=symbol, symbol_table=symbol_table
                )
                if error is not None:
                    errors.append(error)
                    continue

                assert implementation_verify is not None

                if implementation_verify != "":
                    blocks.append(implementation_verify)
        else:
            assert_never(symbol)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            """\
            /// <summary>
            /// Verify the instances of the model classes non-recursively.
            /// </summary>
            /// <remarks>
            /// The methods provided by this class are re-used in the verification
            /// visitors.
            /// </remarks>
            private static class Implementation
            {{
            """
        )
    )
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // private static class Implementation")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_non_recursive_verifier(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the non-recursive verifier which visits the concrete classes."""
    blocks = [
        Stripped("public readonly Verification.Errors Errors;"),
        Stripped(
            textwrap.dedent(
                f"""\
            /// <summary>
            /// Initialize the visitor with the given <paramref name="errors" />.
            ///
            /// The errors observed during the visitation will be appended to
            /// the <paramref name="errors" />.
            /// </summary>
            NonRecursiveVerifier(Verification.Errors errors)
            {{
            {I}Errors = errors;
            }}"""
            )
        ),
        Stripped(
            textwrap.dedent(
                f"""\
            public void Visit(Aas.IClass that, string context)
            {{
            {I}that.Accept(this, context);
            }}"""
            )
        ),
    ]  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if not isinstance(symbol, intermediate.ConcreteClass):
            continue

        cls_name = csharp_naming.class_name(symbol.name)

        blocks.append(
            Stripped(
                textwrap.dedent(
                    f"""\
            /// <summary>
            /// Verify <paramref name="that" /> instance and
            /// append any error to <see cref="Errors" />
            /// where <paramref name="context" /> is used to localize the error.
            /// </summary>
            public void Visit(Aas.{cls_name} that, string context)
            {{
            {I}Implementation.Verify{cls_name}(
            {II}that, context, Errors);
            }}"""
                )
            )
        )

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            f"""\
        /// <summary>
        /// Verify the instances of the model classes non-recursively.
        /// </summary>
        public class NonRecursiveVerifier :
        {I}Visitation.IVisitorWithContext<string>
        {{
        """
        )
    )
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public class NonRecursiveVerifier")

    return Stripped(writer.getvalue()), None


class _RecursionInRecursiveVerifyUnroller(csharp_unrolling.Unroller):
    """Generate the code that unrolls the recursive visits for the given property."""

    #: Symbol to be used to represent references within an AAS
    _ref_association: Final[intermediate.Symbol]

    def __init__(self, ref_association: intermediate.ClassUnion) -> None:
        """Initialize with the given values."""
        self._ref_association = ref_association

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        # We can not recurse visits into a primitive.
        return []

    # noinspection PyUnusedLocal
    def _unroll_our_type_or_ref_annotation(
        self,
        unrollee_expr: str,
        type_annotation: Union[
            intermediate.OurTypeAnnotation, intermediate.RefTypeAnnotation
        ],
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """
        Generate the code for both our atomic type annotations and references.

        We merged :py:method:`._unroll_our_type_annotation` and
        :py:method:`._unroll_ref_type_annotation` together since they differ in only
        which symbol is unrolled over.
        """
        symbol = None  # type: Optional[intermediate.Symbol]
        if isinstance(type_annotation, intermediate.OurTypeAnnotation):
            symbol = type_annotation.symbol
        elif isinstance(type_annotation, intermediate.RefTypeAnnotation):
            symbol = self._ref_association
        else:
            assert_never(type_annotation)

        assert symbol is not None

        if isinstance(
            symbol, (intermediate.Enumeration, intermediate.ConstrainedPrimitive)
        ):
            return []

        assert isinstance(symbol, intermediate.Class), "Exhaustive matching"

        joined_pth = "/".join(path)
        return [
            csharp_unrolling.Node(
                text=textwrap.dedent(
                    f"""\
                if (Errors.Full()) return;
                Visit(
                    {unrollee_expr},
                    ${csharp_common.string_literal(joined_pth)});"""
                ),
                children=[],
            )
        ]

    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        return self._unroll_our_type_or_ref_annotation(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        # NOTE (mristin, 2021-12-19):
        # We need to iterate through the list with the index so that we can construct
        # meaningful paths.

        if item_level > 15:
            index_var = f"i{item_level}"
        else:
            # Use letters i, j, k *etc.* first before we resort to i16, i17 *etc.*
            index_var = chr(ord("i") + item_level)

        children = self.unroll(
            unrollee_expr=f"{unrollee_expr}[{index_var}]",
            type_annotation=type_annotation.items,
            path=path + [f"{{{index_var}}}"],
            item_level=item_level + 1,
            key_value_level=key_value_level,
        )

        if len(children) == 0:
            return []

        text = Stripped(
            f"for(var {index_var} = 0; "
            f"{index_var} < {unrollee_expr}.Count; "
            f"{index_var}++)"
        )

        # Break into lines if too long.
        # This is just a heuristics — we do not consider the actual prefix indention.
        if len(text) > 50:
            text = Stripped(
                textwrap.dedent(
                    f"""\
                    for(
                    {I}var {index_var} = 0;
                    {I}{index_var} < {unrollee_expr}.Count;
                    {I}{index_var}++)"""
                )
            )

        return [csharp_unrolling.Node(text=text, children=children)]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )
        if len(children) > 0:
            return [
                csharp_unrolling.Node(
                    text=f"if ({unrollee_expr} != null)", children=children
                )
            ]
        else:
            return []

    def _unroll_ref_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.RefTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[csharp_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        return self._unroll_our_type_or_ref_annotation(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation,
            path=path,
            item_level=item_level,
            key_value_level=key_value_level,
        )


def _unroll_recursion_in_recursive_verify(
    prop: intermediate.Property, ref_association: intermediate.ClassUnion
) -> Stripped:
    """
    Generate the code for unrolling the recursive visits  for the given property.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """

    prop_name = csharp_naming.property_name(prop.name)

    unroller = _RecursionInRecursiveVerifyUnroller(ref_association=ref_association)
    roots = unroller.unroll(
        unrollee_expr=f"that.{prop_name}",
        type_annotation=prop.type_annotation,
        path=["{context}", prop_name],
        item_level=0,
        key_value_level=0,
    )

    if len(roots) == 0:
        return Stripped("")

    blocks = [csharp_unrolling.render(root) for root in roots]
    return Stripped("\n\n".join(blocks))


# fmt: on
@require(
    lambda cls: not cls.is_implementation_specific,
    "Implementation-specific classes are handled elsewhere",
)
# fmt: off
def _generate_recursive_verifier_visit(
        cls: intermediate.ConcreteClass,
        ref_association: intermediate.ClassUnion
) -> Stripped:
    """
    Generate the ``Visit`` method of the ``RecursiveVerifier`` for the ``cls``.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    cls_name = csharp_naming.class_name(cls.name)

    writer = io.StringIO()
    writer.write(textwrap.dedent(f'''\
        /// <summary>
        /// Verify recursively <paramref name="that" /> instance and
        /// append any error to <see cref="Errors" />
        /// where <paramref name="context" /> is used to localize the error.
        /// </summary>
        public void Visit({cls_name} that, string context)
        {{
        '''))

    blocks = [
        Stripped(textwrap.dedent(f'''\
        Implementation.Verify{cls_name}(
        {I}that, context, Errors);'''))
    ]  # type: List[Stripped]

    # region Unroll

    recursion_ends_here = True
    for prop in cls.properties:
        unrolled_prop_verification = _unroll_recursion_in_recursive_verify(
            prop=prop,
            ref_association=ref_association
        )

        if unrolled_prop_verification != '':
            blocks.append(unrolled_prop_verification)
            recursion_ends_here = False

    if recursion_ends_here:
        blocks.append(Stripped("// The recursion ends here."))
    # endregion

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write('\n\n')

        writer.write(textwrap.indent(block, I))

    writer.write('\n}')
    return Stripped(writer.getvalue())


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_recursive_verifier(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the ``Verifier`` class which visits the classes and verifies them."""
    blocks = [
        Stripped("public readonly Errors Errors;"),
        Stripped(
            textwrap.dedent(
                f"""\
            /// <summary>
            /// Initialize the visitor with the given <paramref name="errors" />.
            ///
            /// The errors observed during the visitation will be appended to
            /// the <paramref name="errors" />.
            /// </summary>
            RecursiveVerifier(Errors errors)
            {{
            {I}Errors = errors;
            }}"""
            )
        ),
        Stripped(
            textwrap.dedent(
                f"""\
            public void Visit(IClass that, string context)
            {{
            {I}that.Accept(this, context);
            }}"""
            )
        ),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for symbol in symbol_table.symbols:
        if not isinstance(symbol, intermediate.ConcreteClass):
            continue

        if symbol.is_implementation_specific:
            visit_key = specific_implementations.ImplementationKey(
                f"Verification/RecursiveVerifier/visit_{symbol.name}.cs"
            )

            implementation = spec_impls.get(visit_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        symbol.parsed.node,
                        f"The implementation snippet is missing for "
                        f"the ``Visit`` method "
                        f"of the ``Verification.RecursiveVerifier`` class: "
                        f"{visit_key}",
                    )
                )
                continue

            blocks.append(implementation)
        else:
            blocks.append(
                _generate_recursive_verifier_visit(
                    cls=symbol, ref_association=symbol_table.ref_association
                )
            )

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        textwrap.dedent(
            f"""\
        /// <summary>
        /// Verify the instances of the model classes recursively.
        /// </summary>
        public class RecursiveVerifier :
        {I}Visitation.IVisitorWithContext<string>
        {{
        """
        )
    )
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public class RecursiveVerifier")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code of the structures based on the symbol table.

    The ``namespace`` defines the AAS C# namespace.
    """
    blocks = [
        csharp_common.WARNING,
        Stripped(
            textwrap.dedent(
                f"""\
            using Regex = System.Text.RegularExpressions.Regex;
            using System.Collections.Generic;  // can't alias
            using System.Collections.ObjectModel;  // can't alias
            using System.Linq;  // can't alias"

            using Aas = {namespace};
            using Visitation = {namespace}.Visitation;"""
            )
        ),
    ]  # type: List[Stripped]

    verification_blocks = []  # type: List[Stripped]
    errors = []  # type: List[Error]

    for implementation_key in [
        specific_implementations.ImplementationKey("Verification/Error.cs"),
        specific_implementations.ImplementationKey("Verification/Errors.cs"),
    ] + [
        specific_implementations.ImplementationKey(f"Verification/{func.name}.cs")
        for func in symbol_table.verification_functions
        if isinstance(func, intermediate.ImplementationSpecificVerification)
    ]:
        implementation = spec_impls.get(implementation_key, None)
        if implementation is None:
            errors.append(Error(None, f"The snippet is missing: {implementation_key}"))
        else:
            verification_blocks.append(implementation)

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            implementation_key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.cs"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        None,
                        f"The snippet for the verification function "
                        f"{verification.name!r} is missing: {implementation_key}",
                    )
                )
            else:
                verification_blocks.append(implementation)

        elif isinstance(verification, intermediate.PatternVerification):
            implementation, error = _transpile_pattern_verification(
                verification=verification
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                verification_blocks.append(implementation)

        else:
            assert_never(verification)

    implementation_class, implementation_class_errors = _generate_implementation_class(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if implementation_class_errors:
        errors.extend(implementation_class_errors)
    else:
        assert implementation_class is not None

        verification_blocks.append(implementation_class)

    non_recursive, non_recursive_errors = _generate_non_recursive_verifier(
        symbol_table=symbol_table
    )

    if non_recursive_errors is not None:
        errors.extend(non_recursive_errors)
    else:
        assert non_recursive is not None

        verification_blocks.append(non_recursive)

    recursive, recursive_errors = _generate_recursive_verifier(
        symbol_table=symbol_table, spec_impls=spec_impls
    )

    if recursive_errors is not None:
        errors.extend(recursive_errors)
    else:
        assert recursive is not None

        verification_blocks.append(recursive)

    if len(errors) > 0:
        return None, errors

    verification_writer = io.StringIO()
    verification_writer.write(f"namespace {namespace}\n{{\n")
    verification_writer.write(f"{I}public static class Verification\n" f"{I}{{\n")

    for i, verification_block in enumerate(verification_blocks):
        if i > 0:
            verification_writer.write("\n\n")

        verification_writer.write(textwrap.indent(verification_block, II))

    verification_writer.write(f"\n{I}}}  // public static class Verification")
    verification_writer.write(f"\n}}  // namespace {namespace}")

    blocks.append(Stripped(verification_writer.getvalue()))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



import re
import textwrap
import unittest
from typing import Tuple, Optional, Sequence

from icontract import ensure

import tests.common
from aas_core_codegen import parse
from aas_core_codegen.common import Error
from aas_core_codegen.intermediate import construction


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def understand_constructor_table(
    source: str,
) -> Tuple[Optional[construction.ConstructorTable], Optional[Error]]:
    """Parse the source and try to understand all the constructors."""
    atok, parse_exception = parse.source_to_atok(source=source)
    assert atok is not None
    assert parse_exception is None, f"{parse_exception=}"

    import_errors = parse.check_expected_imports(atok=atok)
    assert len(import_errors) == 0

    parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    assert error is None, f"{tests.common.most_underlying_messages(error)}"
    assert parsed_symbol_table is not None

    either = construction.understand_all(
        parsed_symbol_table=parsed_symbol_table, atok=atok
    )

    return either


def must_find_item_for(
    constructor_table: construction.ConstructorTable, name: str
) -> Tuple[parse.Class, Sequence[construction.Statement]]:
    """
    Find the constructor statements for the class given with ``identifier``.

    :raise: :py:class:`KeyError` if not found in the table
    """
    for cls, statements in constructor_table.entries():
        if cls.name == name:
            return cls, statements

    raise KeyError(f"Could not find the constructor statements for the class: {name}")


class Test_empty_ok(unittest.TestCase):
    def test_no_constructor(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                pass
                
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constructor_table, error = understand_constructor_table(source=source)
        assert error is None, f"{error=}"
        assert constructor_table is not None

        self.assertEqual(2, len(constructor_table.entries()))

        cls, statements = must_find_item_for(constructor_table, "Something")
        self.assertEqual("Something", cls.name)
        self.assertEqual(0, len(statements))

        cls, statements = must_find_item_for(constructor_table, "Reference")
        self.assertEqual("Reference", cls.name)
        self.assertEqual(0, len(statements))

    def test_pass(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                def __init__(self) -> None:
                    pass
            
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constructor_table, error = understand_constructor_table(source=source)
        assert error is None, f"{error=}"
        assert constructor_table is not None

        self.assertEqual(2, len(constructor_table.entries()))

        _, statements = must_find_item_for(constructor_table, "Something")
        self.assertEqual(0, len(statements))

        _, statements = must_find_item_for(constructor_table, "Reference")
        self.assertEqual(0, len(statements))


class Test_call_to_super_constructor_ok(unittest.TestCase):
    def test_without_arguments(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                def __init__(self) -> None:
                    pass

            class Something(Parent):
                def __init__(self) -> None:
                    Parent.__init__(self)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constructor_table, error = understand_constructor_table(source=source)
        assert error is None, f"{error=}"
        assert constructor_table is not None

        self.assertEqual(3, len(constructor_table.entries()))

        _, statements = must_find_item_for(constructor_table, "Something")
        self.assertEqual(1, len(statements))

        statement = statements[0]
        assert isinstance(statement, construction.CallSuperConstructor)
        self.assertEqual("Parent", statement.super_name)

    def test_with_arguments(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(Parent):
                def __init__(self, a: int, b: int) -> None:
                    Parent.__init__(self, a, b)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constructor_table, error = understand_constructor_table(source=source)
        assert error is None, f"{error=}"
        assert constructor_table is not None

        self.assertEqual(3, len(constructor_table.entries()))

        _, statements = must_find_item_for(constructor_table, "Something")
        self.assertEqual(1, len(statements))

        statement = statements[0]
        assert isinstance(statement, construction.CallSuperConstructor)
        self.assertEqual("Parent", statement.super_name)


class Test_assign_property_ok(unittest.TestCase):
    def test_argument_assignment(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                x: int

                def __init__(self, x: int) -> None:
                    self.x = x
            
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constructor_table, error = understand_constructor_table(source=source)
        assert error is None, f"{error=}"
        assert constructor_table is not None

        _, statements = must_find_item_for(constructor_table, "Something")
        self.assertEqual(1, len(statements))

        statement = statements[0]
        assert isinstance(statement, construction.AssignArgument)
        self.assertEqual("x", statement.name)


class Test_assign_fail(unittest.TestCase):
    def test_multiple_targets_in_assignment(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int
                b: int

                def __init__(self, a: int) -> None:
                    self.a = self.b = a
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected only a single target for property assignment, but got 2 targets",
            tests.common.most_underlying_messages(error),
        )

    def test_tuple_in_assignment_targets(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a, self.b = a, b
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected a property as the target of an assignment, "
            "but got: self.a, self.b",
            tests.common.most_underlying_messages(error),
        )

    def test_variable_instead_of_property_assignment(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    x = a
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected a property as the target of an assignment, but got: x",
            tests.common.most_underlying_messages(error),
        )

    def test_assignment_to_undefined_property(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                def __init__(self, a: int) -> None:
                    self.a = a
            
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "The property has not been previously defined in the class 'Something': a",
            tests.common.most_underlying_messages(error),
        )

    def test_assignment_value_is_not_a_name(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int

                def __init__(self, a: int) -> None:
                    self.a = a + 100
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        error_str = tests.common.most_underlying_messages(error)
        error_str = re.sub(r"Assign\(.*\); ", "Assign(...); ", error_str)

        self.assertEqual(
            "The handling of the constructor statement has not been implemented: "
            "Assign(...); please notify the developers if you really need this feature",
            error_str,
        )

    def test_argument_and_property_name_differ(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int

                def __init__(self, b: int) -> None:
                    self.a = b


            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected the property a to be assigned exactly the argument "
            "with the same name, but got: b",
            tests.common.most_underlying_messages(error),
        )


class Test_call_to_super_constructor_fail(unittest.TestCase):
    def test_super_class_not_a_name(self) -> None:
        source = textwrap.dedent(
            '''\
            @abstract
            class Parent:
                """Represent something abstract."""

                a: int
                b: int

                @require(lambda a: a > 0)
                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(Parent):
                """Represent something concrete.""" 

                def __init__(self, a: int, b: int) -> None:
                    super().__init__(self, a, b)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            '''
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected a super class as a name for a call to super ``__init__``, "
            "but got: super()",
            tests.common.most_underlying_messages(error),
        )

    def test_passed_double_start_keyword_argument(self) -> None:
        source = textwrap.dedent(
            '''\
            @abstract
            class Parent(DBC):
                """Represent something abstract."""

                a: int
                b: int

                @require(lambda a: a > 0)
                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(DBC, Parent):
                """Represent something concrete.""" 

                def __init__(self, a: int, b: int) -> None:
                    Parent.__init__(self, **{'a': a, 'b': b})
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            '''
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected a call to a super ``__init__`` to provide only "
            "explicit keyword arguments, but got a double-star keyword argument",
            tests.common.most_underlying_messages(error),
        )

    def test_calling_constructor_from_a_non_super_class(self) -> None:
        source = textwrap.dedent(
            """\
            class Unrelated:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something:
                def __init__(self, a: int, b: int) -> None:
                    Unrelated.__init__(self, a=a, b=b)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected a super class in the call to a super ``__init__``, "
            "but Something does not inherit from Unrelated",
            tests.common.most_underlying_messages(error),
        )

    def test_super_class_has_no_init(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                pass

            class Something(Parent): 

                def __init__(self, a: int, b: int) -> None:
                    Parent.__init__(self)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "The super class Parent does not define a ``__init__``",
            tests.common.most_underlying_messages(error),
        )

    def test_positional_argument_to_super_init_transformed(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent(DBC):
                a: int

                def __init__(self, a: int) -> None:
                    self.a = a

            class Something(DBC, Parent):
                def __init__(self, a: int) -> None:
                    Parent.__init__(self, a + 100)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected only names in the arguments to super ``__init__``, "
            "but got: a + 100",
            tests.common.most_underlying_messages(error),
        )

    def test_keyword_argument_to_super_init_transformed(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent(DBC):
                a: int

                def __init__(self, a: int) -> None:
                    self.a = a

            class Something(DBC, Parent):
                def __init__(self, a: int) -> None:
                    Parent.__init__(self, a=a + 100)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected only names in the arguments to super ``__init__``, "
            "but got: a + 100",
            tests.common.most_underlying_messages(error),
        )

    def test_too_many_positional_arguments(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(Parent):
                def __init__(self, a: int, b: int, c: int) -> None:
                    Parent.__init__(self, a, b, c)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "The ``Parent.__init__`` expected 3 argument(s), "
            "but the call provides 4 positional argument(s)",
            tests.common.most_underlying_messages(error),
        )

    def test_unexpected_keyword_arguments_supplied_to_super_init(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(DBC, Parent):
                def __init__(self, a: int, b: int, c: int) -> None:
                    Parent.__init__(self, a=a, b=b, c=c)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "The ``Parent.__init__`` does not expect the argument c",
            tests.common.most_underlying_messages(error),
        )

    def test_non_init_names_passed_to_super_init(self) -> None:
        source = textwrap.dedent(
            '''\
            @abstract
            class Parent:
                """Represent something abstract."""

                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(Parent):
                def __init__(self, a: int) -> None:
                    Parent.__init__(self, a, b)
            
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            '''
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Expected all the arguments to ``Parent.__init__`` "
            "to be propagation of the original ``__init__`` arguments, "
            "but the name b is not an argument of ``Something.__init__``",
            tests.common.most_underlying_messages(error),
        )

    def test_arguments_not_passed_as_are(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(DBC, Parent):
                def __init__(self, a: int, y: int) -> None:
                    Parent.__init__(self, a, b=y)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        # This behavior is necessary so that we can easily in-line super-constructors
        # and stack the contracts from the super classes.
        self.assertEqual(
            "Expected the arguments to super ``__init__`` to be passed with "
            "the same names, but the argument b is passed as the name y",
            tests.common.most_underlying_messages(error),
        )

    def test_missing_argument_to_super_init(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Parent:
                a: int
                b: int

                def __init__(self, a: int, b: int) -> None:
                    self.a = a
                    self.b = b

            class Something(Parent):
                def __init__(self, a: int) -> None:
                    Parent.__init__(self, a)
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "The call to ``Parent.__init__`` is missing one or more arguments: b",
            tests.common.most_underlying_messages(error),
        )


class Test_unexpected_statements(unittest.TestCase):
    def test_unexpected_call(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int

                def __init__(self, a: int) -> None:
                    print("something")
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Unexpected call in the body of ``__init__``: print; "
            "only calls to super ``__init__``'s are expected",
            tests.common.most_underlying_messages(error),
        )

    def test_unexpected_expr(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                a: int

                def __init__(self, a: int) -> None:
                    1 + 2
                    
            class Reference:
                pass
            
            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, error = understand_constructor_table(source=source)
        assert error is not None

        self.assertEqual(
            "Unexpected statement in the body of ``__init__``: 1 + 2; "
            "only calls to super ``__init__``'s and property assignments expected",
            tests.common.most_underlying_messages(error),
        )


if __name__ == "__main__":
    unittest.main()



"""Infer constraints representable in common schemas such as JSON Schema or XSD."""

from aas_core_codegen.infer_for_schema import _len, _pattern, _stringify

LenConstraint = _len.LenConstraint
infer_len_constraints_by_class_properties = (
    _len.infer_len_constraints_by_class_properties
)
infer_len_constraint_of_self = _len.infer_len_constraint_of_self

PatternConstraint = _pattern.PatternConstraint
infer_patterns_by_class_properties = _pattern.infer_patterns_by_class_properties
infer_patterns_on_self = _pattern.infer_patterns_on_self
PatternVerificationsByName = _pattern.PatternVerificationsByName
map_pattern_verifications_by_name = _pattern.map_pattern_verifications_by_name

dump = _stringify.dump
dump_len_constraints_by_properties = _stringify.dump_len_constraints_by_properties
dump_patterns = _stringify.dump_patterns
dump_patterns_by_properties = _stringify.dump_patterns_by_properties



"""
Define the parse rules for transforming Python AST to our custom AST.

This module provides translation from a very general Python AST to our more specific,
domain-related syntax.

For comparison, :py:mod:`aas_core_codegen.intermediate` is responsible for
semantic analysis, simplifications and resolution to environments and actual symbols.
"""

import abc
import ast
import os
import pathlib
from typing import Tuple, Optional, List, Mapping, Type, Sequence, Union

from icontract import ensure

from aas_core_codegen.common import Identifier, Error, assert_never
from aas_core_codegen.parse import tree

_AST_COMPARATOR_TO_OURS = {
    ast.Lt: tree.Comparator.LT,
    ast.LtE: tree.Comparator.LE,
    ast.Gt: tree.Comparator.GT,
    ast.GtE: tree.Comparator.GE,
    ast.Eq: tree.Comparator.EQ,
    ast.NotEq: tree.Comparator.NE,
}  # type: Mapping[Type[ast.cmpop], tree.Comparator]

_AST_COMPARATORS = tuple(_AST_COMPARATOR_TO_OURS.keys())


class _Parse(abc.ABC):
    """Define a parse rule from a Python AST node to our custom AST."""

    @abc.abstractmethod
    def matches(self, node: ast.AST) -> bool:
        """Return True if the node can be matched by the rule."""
        raise NotImplementedError()

    @abc.abstractmethod
    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        """Transform the Python node to our custom node."""
        raise NotImplementedError()


class _ParseComparison(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Compare)
            and len(node.ops) == 1
            and isinstance(node.ops[0], _AST_COMPARATORS)
            and len(node.comparators) == 1
        )

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Compare)

        left, error = ast_node_to_our_node(node.left)
        if error is not None:
            return None, error

        assert isinstance(left, tree.Expression), f"{left=}"

        op = _AST_COMPARATOR_TO_OURS[type(node.ops[0])]

        right, error = ast_node_to_our_node(node.comparators[0])
        if error is not None:
            return None, error

        assert isinstance(right, tree.Expression), f"{right=}"

        return tree.Comparison(left=left, op=op, right=right, original_node=node), None


class _ParseCall(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Call)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Call)

        args = []  # type: List[tree.Expression]
        for arg_node in node.args:
            arg, error = ast_node_to_our_node(arg_node)
            if error is not None:
                return None, error

            if not isinstance(arg, tree.Expression):
                return None, Error(
                    arg_node,
                    f"Expected the argument to a call to be an expression, "
                    f"but got: {arg}",
                )

            args.append(arg)

        if len(node.keywords) > 0:
            return None, Error(
                node,
                "Keyword arguments are not supported since "
                "many implementation languages do not support them",
            )

        if isinstance(node.func, ast.Name):
            name, error = ast_node_to_our_node(node.func)
            if error is not None:
                return None, error

            assert name is not None
            assert isinstance(name, tree.Name)

            return (
                tree.FunctionCall(name=name, args=args, original_node=node),
                None,
            )
        else:
            member, error = ast_node_to_our_node(node.func)
            if error is not None:
                return None, error

            assert isinstance(
                member, tree.Member
            ), f"Expected a member for {ast.dump(node.func)=}, but got: {member=}"

            return tree.MethodCall(member=member, args=args, original_node=node), None


class _ParseConstant(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Constant) and isinstance(
            node.value, (bool, int, float, str)
        )

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Constant) and isinstance(
            node.value, (bool, int, float, str)
        )
        return tree.Constant(value=node.value, original_node=node), None


class _ParseImplication(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return (
            isinstance(node, ast.BoolOp)
            and isinstance(node.op, ast.Or)
            and len(node.values) == 2
            and isinstance(node.values[0], ast.UnaryOp)
            and isinstance(node.values[0].op, ast.Not)
        )

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert (
            isinstance(node, ast.BoolOp)
            and isinstance(node.op, ast.Or)
            and len(node.values) == 2
            and isinstance(node.values[0], ast.UnaryOp)
            and isinstance(node.values[0].op, ast.Not)
        )

        antecedent, error = ast_node_to_our_node(node.values[0].operand)
        if error is not None:
            return None, error

        assert isinstance(antecedent, tree.Expression), f"{antecedent=}"

        consequent, error = ast_node_to_our_node(node.values[1])
        if error is not None:
            return None, error

        assert isinstance(consequent, tree.Expression), f"{consequent=}"

        return (
            tree.Implication(
                antecedent=antecedent, consequent=consequent, original_node=node
            ),
            None,
        )


class _ParseMember(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Attribute)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Attribute)

        instance, error = ast_node_to_our_node(node.value)
        if error is not None:
            return None, error

        assert isinstance(instance, tree.Expression), f"{instance=}"

        return (
            tree.Member(
                instance=instance, name=Identifier(node.attr), original_node=node
            ),
            None,
        )


class _ParseName(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Name)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Name)

        return tree.Name(identifier=Identifier(node.id), original_node=node), None


class _ParseIsNoneOrIsNotNone(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Compare)
            and len(node.ops) == 1
            and isinstance(node.ops[0], (ast.Is, ast.IsNot))
            and len(node.comparators) == 1
            and isinstance(node.comparators[0], ast.Constant)
            and node.comparators[0].value is None
        )

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert (
            isinstance(node, ast.Compare)
            and len(node.ops) == 1
            and isinstance(node.ops[0], (ast.Is, ast.IsNot))
            and len(node.comparators) == 1
            and isinstance(node.comparators[0], ast.Constant)
            and node.comparators[0].value is None
        )

        value, error = ast_node_to_our_node(node.left)
        if error is not None:
            return None, error

        assert value is not None
        assert isinstance(value, tree.Expression), f"{value=}"

        if isinstance(node.ops[0], ast.Is):
            return tree.IsNone(value=value, original_node=node), None
        elif isinstance(node.ops[0], ast.IsNot):
            return tree.IsNotNone(value=value, original_node=node), None
        else:
            raise AssertionError(f"Unexpected: {node.ops[0]=}")


class _ParseAndOrOr(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.BoolOp) and isinstance(node.op, (ast.And, ast.Or))

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.BoolOp) and isinstance(node.op, (ast.And, ast.Or))

        values = []  # type: List[tree.Expression]
        for value_node in node.values:
            value, error = ast_node_to_our_node(value_node)
            if error is not None:
                return None, error

            assert value is not None
            assert isinstance(value, tree.Expression), f"{value=}"

            values.append(value)

        if isinstance(node.op, ast.And):
            return tree.And(values=values, original_node=node), None
        elif isinstance(node.op, ast.Or):
            return tree.Or(values=values, original_node=node), None
        else:
            raise AssertionError(f"Unexpected: {node.op=}")


class _ParseExpressionWithDeclaration(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return (
            isinstance(node, ast.Subscript)
            and isinstance(node.value, ast.Tuple)
            and len(node.value.elts) == 2
            and isinstance(node.value.elts[0], ast.NamedExpr)
            and isinstance(node.slice, ast.Index)
            and isinstance(node.slice.value, ast.Constant)
            and node.slice.value.value == 1
        )

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert (
            isinstance(node, ast.Subscript)
            and isinstance(node.value, ast.Tuple)
            and len(node.value.elts) == 2
            and isinstance(node.value.elts[0], ast.NamedExpr)
            and isinstance(node.slice, ast.Index)
            and isinstance(node.slice.value, ast.Constant)
            and node.slice.value.value == 1
        )

        declaration, error = ast_node_to_our_node(node.value.elts[0])
        if error is not None:
            return None, error

        assert isinstance(declaration, tree.Declaration), f"{declaration=}"

        expression, error = ast_node_to_our_node(node.value.elts[1])
        if error is not None:
            return None, error

        assert isinstance(expression, tree.Expression), f"{expression=}"

        return (
            tree.ExpressionWithDeclarations(
                declarations=[declaration], expression=expression, original_node=node
            ),
            None,
        )


class _ParseDeclaration(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.NamedExpr) and isinstance(node.target, ast.Name)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.NamedExpr) and isinstance(node.target, ast.Name)

        value, error = ast_node_to_our_node(node.value)
        if error is not None:
            return None, error

        assert value is not None
        assert isinstance(value, tree.Expression), f"{value=}"

        return (
            tree.Declaration(
                identifier=Identifier(node.target.id), value=value, original_node=node
            ),
            None,
        )


class _ParseExpression(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Expr)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Expr)

        value, error = ast_node_to_our_node(node.value)
        if error is not None:
            return None, error

        assert value is not None

        return value, None


class _ParseJoinedStr(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.JoinedStr)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.JoinedStr)

        values = []  # type: List[Union[str, tree.FormattedValue]]

        assert isinstance(node, ast.JoinedStr)
        for value_node in node.values:
            if isinstance(value_node, ast.Constant):
                if not isinstance(value_node.value, str):
                    return None, Error(
                        value_node,
                        "Unexpected non-string constant in the joined string",
                    )

                values.append(value_node.value)

            elif isinstance(value_node, ast.FormattedValue):
                if value_node.conversion != -1:
                    return None, Error(
                        value_node,
                        f"We do not support any conversions at the moment. "
                        f"Expected -1, but got conversion: {value_node.conversion}",
                    )

                if value_node.format_spec is not None:
                    return None, Error(
                        value_node,
                        f"We do not support any format spec at the moment. "
                        f"Expected None, but got format spec: {value_node.format_spec}",
                    )

                # noinspection PyTypeChecker
                value, error = ast_node_to_our_node(node=value_node.value)
                if error is not None:
                    return None, error

                assert value is not None

                assert isinstance(value, tree.Expression)
                values.append(
                    tree.FormattedValue(value=value, original_node=value_node)
                )

            elif isinstance(value_node, ast.expr):
                return None, Error(
                    value_node,
                    f"We do not know how to parse the Python ast.expr "
                    f"as part of the JoinedStr.values: {ast.dump(value_node)}; "
                    f"please notify the developers.",
                )
            else:
                assert_never(value_node)

        return tree.JoinedStr(values=values, original_node=node), None


class _ParseAssignment(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Assign) and len(node.targets) == 1

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Assign) and len(node.targets) == 1

        target, error = ast_node_to_our_node(node.targets[0])
        if error is not None:
            return None, error

        assert target is not None
        assert isinstance(target, tree.Expression), f"{target=}"

        value, error = ast_node_to_our_node(node.value)
        if error is not None:
            return None, error

        assert value is not None
        assert isinstance(value, tree.Expression), f"{value=}"

        return (
            tree.Assignment(target=target, value=value, original_node=node),
            None,
        )


class _ParseReturn(_Parse):
    def matches(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Return)

    def transform(self, node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
        assert isinstance(node, ast.Return)

        if node.value is None:
            return tree.Return(value=None, original_node=node), None

        value, error = ast_node_to_our_node(node.value)
        if error is not None:
            return None, error

        assert value is not None
        assert isinstance(value, tree.Expression), f"{value=}"

        return tree.Return(value=value, original_node=node), None


_CHAIN_OF_RULES = [
    _ParseComparison(),
    _ParseCall(),
    _ParseConstant(),
    _ParseImplication(),
    _ParseMember(),
    _ParseName(),
    _ParseIsNoneOrIsNotNone(),
    _ParseAndOrOr(),
    _ParseExpressionWithDeclaration(),
    _ParseDeclaration(),
    _ParseExpression(),
    _ParseJoinedStr(),
    _ParseAssignment(),
    _ParseReturn(),
]  # type: Sequence[_Parse]


def _assert_chains_follow_file_structure() -> None:
    """
    Make sure that the chains of command follow the structure of the module.

    This check is necessary so that the rules can be directly followed in the source
    code. Otherwise, it is very hard to follow the chain if it differs from the order
    in which the classes are defined.
    """
    this_file = pathlib.Path(os.path.realpath(__file__))

    # If we are in an environment where we can not load the file, skip this assertion.
    # This is the case, for example, in a pyinstaller package.
    if not this_file.exists():
        return

    root = ast.parse(
        source=this_file.read_text(encoding="utf-8"), filename=str(this_file)
    )
    assert isinstance(root, ast.Module)

    expected_parse_names = [
        stmt.name
        for stmt in root.body
        if (
            isinstance(stmt, ast.ClassDef)
            and stmt.name.startswith("_Parse")
            and stmt.name != "_Parse"
        )
    ]  # type: List[str]

    parse_names_in_chain = [parse.__class__.__name__ for parse in _CHAIN_OF_RULES]

    assert (
        expected_parse_names == parse_names_in_chain
    ), f"{expected_parse_names=} != {parse_names_in_chain=}"


_assert_chains_follow_file_structure()


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def ast_node_to_our_node(node: ast.AST) -> Tuple[Optional[tree.Node], Optional[Error]]:
    """
    Parse the Python AST node into our custom AST.

    For example, this function is used to parse contract conditions into
    our representation which is later easier for processing.
    """
    for parse_rule in _CHAIN_OF_RULES:
        # NOTE(mristin, 2021-10-08):
        # Please leave the variables as they are to facilitate the eventual debugging
        # even though a more succinct code structure lures you.

        matches = parse_rule.matches(node)
        if matches:
            result, error = parse_rule.transform(node)

            if error is not None:
                return None, error

            return result, None

    return None, Error(
        node,
        f"The code matched no pattern for transpilation "
        f"at the parse stage: {ast.dump(node)}",
    )



import textwrap
import unittest
from typing import Tuple, Optional, MutableMapping, List

from icontract import ensure

import tests.common
from aas_core_codegen import intermediate, infer_for_schema
from aas_core_codegen.common import Identifier, Error


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_constraints_on_self_of_class_something(
    source: str,
) -> Tuple[Optional[infer_for_schema.LenConstraint], Optional[List[Error]]]:
    """Translate the ``source`` into inferred constraints on the class ``Something``."""
    symbol_table, error = tests.common.translate_source_to_intermediate(source=source)
    assert error is None, tests.common.most_underlying_messages(error)
    assert symbol_table is not None
    symbol = symbol_table.must_find(Identifier("Something"))
    assert isinstance(symbol, intermediate.ConstrainedPrimitive)

    result = infer_for_schema.infer_len_constraint_of_self(constrained_primitive=symbol)

    return result


class Test_expected(unittest.TestCase):
    def test_no_constraints(self) -> None:
        source = textwrap.dedent(
            """\
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=None,
                  max_value=None)"""
            ),
            text,
        )

    def test_min_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 < len(self))
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=11,
                  max_value=None)"""
            ),
            text,
        )

    def test_min_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) > 10)
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=11,
                  max_value=None)"""
            ),
            text,
        )

    def test_max_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) < 10)
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=None,
                  max_value=9)"""
            ),
            text,
        )

    def test_max_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 > len(self))
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=None,
                  max_value=9)"""
            ),
            text,
        )

    def test_exact_value_constant_left(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: 10 == len(self))
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                        LenConstraint(
                          min_value=10,
                          max_value=10)"""
            ),
            text,
        )

    def test_exact_value_constant_right(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) == 10)
            class Something(str):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=10,
                  max_value=10)"""
            ),
            text,
        )

    def test_inheritance(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) > 3)
            class Parent(str):
                pass


            @invariant(lambda self: len(self) > 5) 
            class Something(Parent):
                pass 


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        constraint, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is None, tests.common.most_underlying_messages(errors)
        assert constraint is not None

        # NOTE (mristin, 2022-01-02):
        # We infer only the constraints as specified in the class itself, and
        # ignore the constraints of the ancestors in *this particular kind of
        # inference*.

        text = infer_for_schema.dump(constraint)
        self.assertEqual(
            textwrap.dedent(
                """\
                LenConstraint(
                  min_value=6,
                  max_value=None)"""
            ),
            text,
        )


class Test_unexpected(unittest.TestCase):
    def test_conflicting_min_and_max(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) > 10)
            @invariant(lambda self: len(self) < 3) 
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "There are conflicting invariants on the length: "
            "the minimum length, 11, contradicts the maximum length 2.",
            tests.common.most_underlying_messages(errors),
        )

    def test_conflicting_min_and_exact(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) > 10)
            @invariant(lambda self: len(self) == 3) 
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "There are conflicting invariants on the length: "
            "the minimum length, 11, contradicts the exactly expected length 3.",
            tests.common.most_underlying_messages(errors),
        )

    def test_conflicting_max_and_exact(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) < 10)
            @invariant(lambda self: len(self) == 30) 
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        _, errors = infer_constraints_on_self_of_class_something(source=source)

        assert errors is not None
        self.assertEqual(
            "There are conflicting invariants on the length: "
            "the maximum length, 9, contradicts the exactly expected length 30.",
            tests.common.most_underlying_messages(errors),
        )


if __name__ == "__main__":
    unittest.main()



import textwrap
import unittest
from typing import MutableMapping, List

import tests.common
from aas_core_codegen import intermediate, infer_for_schema
from aas_core_codegen.common import Identifier


def infer_patterns_by_properties_of_class_something(
    source: str,
) -> MutableMapping[intermediate.Property, List[infer_for_schema.PatternConstraint]]:
    """Translate the ``source`` into inferred constraints of the class ``Something``."""
    symbol_table, error = tests.common.translate_source_to_intermediate(source=source)
    assert error is None, tests.common.most_underlying_messages(error)
    assert symbol_table is not None
    symbol = symbol_table.must_find(Identifier("Something"))
    assert isinstance(symbol, intermediate.Class)

    pattern_verifications_by_name = infer_for_schema.map_pattern_verifications_by_name(
        verifications=symbol_table.verification_functions
    )

    return infer_for_schema.infer_patterns_by_class_properties(
        cls=symbol, pattern_verifications_by_name=pattern_verifications_by_name
    )


class Test_expected(unittest.TestCase):
    def test_no_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props = infer_patterns_by_properties_of_class_something(source=source)

        assert by_props is not None

        text = infer_for_schema.dump_patterns_by_properties(by_props)
        self.assertEqual("{}", text)

    def test_single_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                prefix = "something"
                return match(f"{prefix}-[a-zA-Z]+", text) is not None

            
            @invariant(lambda self: is_something(self.some_property))
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props = infer_patterns_by_properties_of_class_something(source=source)

        assert by_props is not None

        text = infer_for_schema.dump_patterns_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                  'some_property':
                  [
                    PatternConstraint(
                      pattern='something-[a-zA-Z]+')
                  ]
                }"""
            ),
            text,
        )

    def test_two_patterns(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None


            @invariant(lambda self: is_acme(self.some_property))
            @invariant(lambda self: is_something(self.some_property))
            class Something:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props = infer_patterns_by_properties_of_class_something(source=source)

        assert by_props is not None

        text = infer_for_schema.dump_patterns_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                  'some_property':
                  [
                    PatternConstraint(
                      pattern='something-[a-zA-Z]+'),
                    PatternConstraint(
                      pattern='.*acme.*')
                  ]
                }"""
            ),
            text,
        )

    def test_conditioned_on_property(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @invariant(
                lambda self: 
                not (self.some_property is not None) 
                or  is_something(self.some_property)
            )
            class Something:
                some_property: Optional[str]
            
                def __init__(self, some_property: Optional[str] = None) -> None:
                    self.some_property = some_property


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props = infer_patterns_by_properties_of_class_something(source=source)

        assert by_props is not None

        # NOTE (mristin, 2022-01-02):
        # We infer only the constraints as specified in the class itself, and
        # ignore the constraints of the ancestors in *this particular kind of
        # inference*.

        text = infer_for_schema.dump_patterns_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                  'some_property':
                  [
                    PatternConstraint(
                      pattern='something-[a-zA-Z]+')
                  ]
                }"""
            ),
            text,
        )

    def test_inheritance(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None

            @invariant(lambda self: is_something(self.some_property))
            class Parent:
                some_property: str

                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property


            @invariant(lambda self: is_acme(self.some_property))
            class Something(Parent):
                def __init__(self, some_property: str) -> None:
                    Parent.__init__(self, some_property=some_property)


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        by_props = infer_patterns_by_properties_of_class_something(source=source)

        assert by_props is not None

        # NOTE (mristin, 2022-01-02):
        # We infer only the constraints as specified in the class itself, and
        # ignore the constraints of the ancestors in *this particular kind of
        # inference*.

        text = infer_for_schema.dump_patterns_by_properties(by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                  'some_property':
                  [
                    PatternConstraint(
                      pattern='.*acme.*')
                  ]
                }"""
            ),
            text,
        )


if __name__ == "__main__":
    unittest.main()
